@@ -6,25 +6,38 @@
 # GNU General Public License version 2, incorporated herein by reference.
 
 import os
+import re
 import gtk
 import gobject
 import pango
 import Queue
 
-from mercurial import cmdutil, util, patch, mdiff
+from mercurial import cmdutil, util, patch, mdiff, error
 
 from tortoisehg.util.i18n import _
-from tortoisehg.util import shlib, hglib
+from tortoisehg.util import shlib, hglib, paths
 
 from tortoisehg.hgtk import csinfo, gdialog, gtklib, hgcmd, statusbar
 
-class ChangeSet(gdialog.GDialog):
+class ChangeSet(gdialog.GWindow):
     'GTK+ based dialog for displaying repository logs'
     def __init__(self, ui, repo, cwd, pats, opts, stbar=None):
-        gdialog.GDialog.__init__(self, ui, repo, cwd, pats, opts)
+        gdialog.GWindow.__init__(self, ui, repo, cwd, pats, opts)
         self.stbar = stbar
         self.glog_parent = None
         self.bfile = None
+        self.colorstyle = repo.ui.config('tortoisehg', 'diffcolorstyle')
+
+        # initialize changeset/issue tracker link regex and dict
+        csmatch = r'(\b[0-9a-f]{12}(?:[0-9a-f]{28})?\b)'
+        httpmatch = r'(\b(http|https)://([-A-Za-z0-9+&@#/%?=~_()|!:,.;]*[-A-Za-z0-9+&@#/%=~_()|]))'
+        issue = repo.ui.config('tortoisehg', 'issue.regex')
+        if issue:
+            regexp = r'%s|%s|(%s)' % (csmatch, httpmatch, issue)
+        else:
+            regexp = r'%s|%s' % (csmatch, httpmatch)
+        self.bodyre = re.compile(regexp)
+        self.issuedict = dict()
 
     def get_title(self):
         title = _('%s changeset ') % self.get_reponame()
@@ -52,12 +65,12 @@ class ChangeSet(gdialog.GDialog):
         self.load_details(self.repo.changelog.rev(node0))
 
     def save_settings(self):
-        settings = gdialog.GDialog.save_settings(self)
+        settings = gdialog.GWindow.save_settings(self)
         settings['changeset'] = self._hpaned.get_position()
         return settings
 
     def load_settings(self, settings):
-        gdialog.GDialog.load_settings(self, settings)
+        gdialog.GWindow.load_settings(self, settings)
         if settings and 'changeset' in settings:
             self._setting_hpos = settings['changeset']
         else:
@@ -92,8 +105,8 @@ class ChangeSet(gdialog.GDialog):
             if len(parents) == 2:
                 # deferred adding of parent check button
                 if not self.parent_button.parent:
-                    self.parent_box.pack_start(gtk.HSeparator(), False, False)
                     self.parent_box.pack_start(self.parent_button, False, False)
+                    self.parent_box.pack_start(gtk.HSeparator(), False, False)
                     self.parent_box.show_all()
 
                 # show parent box
@@ -124,6 +137,19 @@ class ChangeSet(gdialog.GDialog):
         else:
             parent = self.repo[-1]
 
+        oldother = hasattr(self, 'otherparent') and self.otherparent or None
+        self.otherparent = len(parents) == 2 and self.diff_other_parent()
+
+        # refresh merge row without graph redrawing
+        if self.graphview:
+            gview = self.graphview
+            if oldother:
+                gview.model.clear_parents()
+            elif self.otherparent:
+                gview.model.set_parent(ctx.rev(), parent)
+            gview.hide()
+            gview.show()
+
         # update dialog title
         self.set_title(title)
 
@@ -157,8 +183,10 @@ class ChangeSet(gdialog.GDialog):
         self.curnodes = (parent, ctx.node())
         if selrow is not None:
             self._filesel.select_path((selrow,))
+            self._filelist_tree.set_cursor((selrow,))
         elif len(self._filelist) > 1:
             self._filesel.select_path((1,))
+            self._filelist_tree.set_cursor((1,))
         else:
             self._filesel.select_path((0,))
 
@@ -261,7 +289,30 @@ class ChangeSet(gdialog.GDialog):
         buf = self._buffer
         buf.set_text('')
         eob = buf.get_end_iter()
-        buf.insert(eob, desc.rstrip('\n\r') + '\n\n')
+        desc = desc.rstrip('\n\r')
+
+        pos = 0
+        self.issuedict.clear()
+        for m in self.bodyre.finditer(desc):
+            a, b = m.span()
+            if a > pos:
+                buf.insert(eob, desc[pos:a])
+                pos = b
+            groups = m.groups()
+            if groups[0]:
+                link = groups[0]
+                buf.insert_with_tags_by_name(eob, link, 'csetlink')
+            elif groups[1]:
+                link = groups[1]
+                buf.insert_with_tags_by_name(eob, link, 'urllink')
+            else:
+                link = groups[4]
+                if len(groups) > 4:
+                    self.issuedict[link] = groups[4:]
+                buf.insert_with_tags_by_name(eob, link, 'issuelink')
+        if pos < len(desc):
+            buf.insert(eob, desc[pos:])
+        buf.insert(eob, '\n\n')
 
     def append_diff(self, wfile):
         if not wfile:
@@ -274,7 +325,7 @@ class ChangeSet(gdialog.GDialog):
 
         try:
             fctx = self.repo[rev].filectx(wfile)
-        except hglib.LookupError:
+        except error.LookupError:
             fctx = None
         if fctx and fctx.size() > hglib.getmaxdiffsize(self.repo.ui):
             lines = ['diff',
@@ -286,7 +337,7 @@ class ChangeSet(gdialog.GDialog):
             try:
                 for s in patch.diff(self.repo, n1, n2, match=m, opts=opts):
                     lines.extend(s.splitlines())
-            except (hglib.RepoLookupError, hglib.RepoError, hglib.LookupError), e:
+            except (error.RepoLookupError, error.RepoError, error.LookupError), e:
                 err = _('Repository Error:  %s, refresh suggested') % str(e)
                 lines = ['diff', '', err]
         tags, lines = self.prepare_diff(lines, offset, wfile)
@@ -358,9 +409,9 @@ class ChangeSet(gdialog.GDialog):
         offset += len(txt.decode('utf-8'))
         for l1 in difflines[1:]:
             l = hglib.toutf(l1)
-            if l.startswith('+++'):
+            if l.startswith('--- '):
                 continue
-            if l.startswith('---'):
+            if l.startswith('+++ '):
                 continue
             if l.startswith('@@'):
                 tag = 'blue'
@@ -388,28 +439,23 @@ class ChangeSet(gdialog.GDialog):
             self.load_details(revnum)
 
     def file_context_menu(self):
-        def create_menu(label, callback):
-            menuitem = gtk.MenuItem(label, True)
-            menuitem.connect('activate', callback)
-            menuitem.set_border_width(1)
-            return menuitem
+        m = gtklib.MenuBuilder()
+        m.append(_('_Visual Diff'), self.diff_file_rev,
+                 gtk.STOCK_JUSTIFY_FILL)
+        m.append(_('Diff to _local'), self.diff_to_local)
+        m.append_sep()
+        m.append(_('_View at Revision'), self.view_file_rev, gtk.STOCK_EDIT)
+        self.msave = m.append(_('_Save at Revision...'),
+                              self.save_file_rev, gtk.STOCK_SAVE)
+        m.append_sep()
+        m.append(_('_File History'), self.file_history, 'menulog.ico')
+        self.ann_menu = m.append(_('_Annotate File'), self.ann_file,
+                                 'menublame.ico')
+        m.append_sep()
+        m.append(_('_Revert File Contents'), self.revert_file,
+                 gtk.STOCK_MEDIA_REWIND)
 
-        menu = gtklib.MenuItems()
-        menu.append(create_menu(_('_Visual Diff'), self.diff_file_rev))
-        menu.append(create_menu(_('Diff to _local'), self.diff_to_local))
-        menu.append_sep()
-        menu.append(create_menu(_('_View at Revision'), self.view_file_rev))
-        self.save_menu = create_menu(_('_Save at Revision...'),
-            self.save_file_rev)
-        menu.append(self.save_menu)
-        menu.append_sep()
-        menu.append(create_menu(_('_File History'), self.file_history))
-        self.ann_menu = create_menu(_('_Annotate File'), self.ann_file)
-        menu.append(self.ann_menu)
-        menu.append_sep()
-        menu.append(create_menu(_('_Revert File Contents'), self.revert_file))
-
-        menu = menu.create_menu()
+        menu = m.build()
         menu.show_all()
         return menu
 
@@ -478,14 +524,14 @@ class ChangeSet(gdialog.GDialog):
                         branch = cctx.branch()
                     children.append(revline_data(cctx, branch=branch))
                 return children
-            elif item == 'transplant':
-                ts = widget.get_data('transplant', usepreset=True)
+            elif item in ('transplant', 'p4', 'svn'):
+                ts = widget.get_data(item, usepreset=True)
                 if not ts:
                     return None
                 try:
                     tctx = self.repo[ts]
                     return revline_data(tctx)
-                except (hglib.LookupError, hglib.RepoLookupError, hglib.RepoError):
+                except (error.LookupError, error.RepoLookupError, error.RepoError):
                     return ts
             elif item == 'patch':
                 if hasattr(ctx, '_patchname'):
@@ -516,7 +562,7 @@ class ChangeSet(gdialog.GDialog):
                     if branch:
                         return '%s - %s %s' % (revnum, branch, summary)
                     return '%s - %s' % (revnum, summary)
-            if item in ('cset', 'transplant', 'patch'):
+            if item in ('cset', 'transplant', 'patch', 'p4', 'svn'):
                 if isinstance(value, basestring):
                     return revid_markup(value)
                 return revline_markup(*value)
@@ -532,7 +578,7 @@ class ChangeSet(gdialog.GDialog):
         def widget_func(widget, item, markups):
             def linkwidget(revnum, revid, summary, highlight=None, branch=None):
                 # revision label
-                opts = dict(underline='single', foreground='#0000FF')
+                opts = dict(underline='single', color='blue')
                 if highlight:
                     opts['weight'] = 'bold'
                 rev = '%s (%s)' % (gtklib.markup(revnum, **opts),
@@ -545,7 +591,7 @@ class ChangeSet(gdialog.GDialog):
                 sum = gtklib.markup(summary)
                 if branch:
                     sum = gtklib.markup(branch, color='black',
-                        background='#aaffaa') + ' ' + sum
+                        background=gtklib.PGREEN) + ' ' + sum
                 sumlabel = gtk.Label()
                 sumlabel.set_markup(sum)
                 sumlabel.set_selectable(True)
@@ -572,7 +618,7 @@ class ChangeSet(gdialog.GDialog):
                                markup=markup_func, widget=widget_func)
         self.csetstyle = csinfo.panelstyle(contents=('cset', 'branch',
                                 'user', 'dateage', 'parents', 'children',
-                                'tags', 'transplant'), selectable=True)
+                                'tags', 'transplant', 'p4', 'svn'), selectable=True)
         self.patchstyle = csinfo.panelstyle(contents=('patch', 'branch',
                                  'user', 'dateage', 'parents'),
                                  selectable=True)
@@ -598,7 +644,7 @@ class ChangeSet(gdialog.GDialog):
         details_text.set_wrap_mode(gtk.WRAP_NONE)
         details_text.connect('populate-popup', self.add_to_popup)
         details_text.set_editable(False)
-        details_text.modify_font(pango.FontDescription(self.fontcomment))
+        details_text.modify_font(self.fonts['comment'])
         details_box.pack_start(details_text)
 
         self._buffer = gtk.TextBuffer()
@@ -608,6 +654,7 @@ class ChangeSet(gdialog.GDialog):
 
         ## file list
         filelist_tree = gtk.TreeView()
+        filelist_tree.set_headers_visible(False)
         filesel = filelist_tree.get_selection()
         filesel.connect('changed', self.filelist_rowchanged)
         self._filesel = filesel
@@ -616,7 +663,8 @@ class ChangeSet(gdialog.GDialog):
         filelist_tree.connect('popup-menu', self.file_popup_menu)
         filelist_tree.connect('row-activated', self.file_row_act)
         filelist_tree.set_search_equal_func(self.search_filelist)
-        filelist_tree.modify_font(pango.FontDescription(self.fontlist))
+        filelist_tree.modify_font(self.fonts['list'])
+        self._filelist_tree = filelist_tree
 
         accelgroup = gtk.AccelGroup()
         if self.glog_parent:
@@ -629,28 +677,72 @@ class ChangeSet(gdialog.GDialog):
                         modifier, gtk.ACCEL_VISIBLE)
         filelist_tree.connect('thg-diff', self.thgdiff)
 
+        def scroll_details(widget, direction=gtk.SCROLL_PAGE_DOWN):
+            self.diffscroller.emit("scroll-child", direction, False)
+
+        # signal, accelerator key, handler, (parameters,)
+        status_accelerators = [
+            ('status-scroll-down', 'bracketright', scroll_details,
+             (gtk.SCROLL_PAGE_DOWN,)),
+            ('status-scroll-up', 'bracketleft', scroll_details,
+             (gtk.SCROLL_PAGE_UP,)),
+            ('status-next-file', 'period', gtklib.move_treeview_selection,
+             (filelist_tree, 1)),
+            ('status-previous-file', 'comma', gtklib.move_treeview_selection,
+             (filelist_tree, -1)),
+        ]
+        
+        for signal, accelerator, handler, param in status_accelerators:
+            root = self.glog_parent or self
+            gtklib.add_accelerator(root, signal, accelgroup,
+                                   mod + accelerator)
+            root.connect(signal, handler, *param)
+
         self._filelist = gtk.ListStore(
                 gobject.TYPE_STRING,   # MAR status
                 gobject.TYPE_STRING,   # filename (utf-8 encoded)
                 gobject.TYPE_STRING,   # filename
                 )
         filelist_tree.set_model(self._filelist)
-        column = gtk.TreeViewColumn(_('Stat'), gtk.CellRendererText(), text=0)
-        filelist_tree.append_column(column)
-        column = gtk.TreeViewColumn(_('Files'), gtk.CellRendererText(), text=1)
+
+        column = gtk.TreeViewColumn()
         filelist_tree.append_column(column)
 
+        iconcell = gtk.CellRendererPixbuf()
+        filecell = gtk.CellRendererText()
+
+        column.pack_start(iconcell, expand=False)
+        column.pack_start(filecell, expand=False)
+        column.add_attribute(filecell, 'text', 1)
+
+        size = gtk.ICON_SIZE_SMALL_TOOLBAR
+        addedpixbuf = gtklib.get_icon_pixbuf('fileadd.ico', size)
+        removedpixbuf = gtklib.get_icon_pixbuf('filedelete.ico', size)
+        modifiedpixbuf = gtklib.get_icon_pixbuf('filemodify.ico', size)
+
+        def cell_seticon(column, cell, model, iter):
+            state = model.get_value(iter, 0)
+            pixbuf = None
+            if state == 'A':
+                pixbuf = addedpixbuf
+            elif state == 'R':
+                pixbuf = removedpixbuf
+            elif state == 'M':
+                pixbuf = modifiedpixbuf
+            cell.set_property('pixbuf', pixbuf)
+
+        column.set_cell_data_func(iconcell, cell_seticon)
+
         list_frame = gtk.Frame()
         list_frame.set_shadow_type(gtk.SHADOW_ETCHED_IN)
         scroller = gtk.ScrolledWindow()
         scroller.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
         scroller.add(filelist_tree)
         flbox = gtk.VBox()
-        flbox.pack_start(scroller)
         list_frame.add(flbox)
-
         self.parent_box = gtk.VBox()
         flbox.pack_start(self.parent_box, False, False)
+        flbox.pack_start(scroller)
 
         btn = gtk.CheckButton(_('Diff to second Parent'))
         btn.connect('toggled', self.parent_toggled)
@@ -700,16 +792,85 @@ class ChangeSet(gdialog.GDialog):
 
         tag_table = self._buffer.get_tag_table()
 
-        tag_table.add(make_texttag('diff', font=self.fontdiff))
+        tag_table.add(make_texttag('diff', font=self.rawfonts['fontdiff']))
         tag_table.add(make_texttag('blue', foreground='blue'))
-        tag_table.add(make_texttag('red', foreground='red'))
-        tag_table.add(make_texttag('green', foreground='darkgreen'))
+        if self.colorstyle == 'background':
+            tag_table.add(make_texttag('red',
+                                       paragraph_background=gtklib.PRED))
+            tag_table.add(make_texttag('green',
+                                       paragraph_background=gtklib.PGREEN))
+        elif self.colorstyle == 'none':
+            tag_table.add(make_texttag('red'))
+            tag_table.add(make_texttag('green'))
+        else:
+            tag_table.add(make_texttag('red', foreground=gtklib.DRED))
+            tag_table.add(make_texttag('green', foreground=gtklib.DGREEN))
         tag_table.add(make_texttag('black', foreground='black'))
         tag_table.add(make_texttag('greybg',
                                    paragraph_background='grey',
                                    weight=pango.WEIGHT_BOLD))
         tag_table.add(make_texttag('yellowbg', background='yellow'))
 
+        issuelink_tag = make_texttag('issuelink', foreground='blue',
+                                     underline=pango.UNDERLINE_SINGLE)
+        issuelink_tag.connect('event', self.issuelink_event)
+        tag_table.add(issuelink_tag)
+        urllink_tag = make_texttag('urllink', foreground='blue',
+                                     underline=pango.UNDERLINE_SINGLE)
+        urllink_tag.connect('event', self.urllink_event)
+        tag_table.add(urllink_tag)
+        csetlink_tag = make_texttag('csetlink', foreground='blue',
+                                    underline=pango.UNDERLINE_SINGLE)
+        csetlink_tag.connect('event', self.csetlink_event)
+        tag_table.add(csetlink_tag)
+
+    def urllink_event(self, tag, widget, event, liter):
+        if event.type != gtk.gdk.BUTTON_RELEASE:
+            return
+        text = self.get_link_text(tag, widget, liter)
+        shlib.browse_url(text)
+
+    def issuelink_event(self, tag, widget, event, liter):
+        if event.type != gtk.gdk.BUTTON_RELEASE:
+            return
+        text = self.get_link_text(tag, widget, liter)
+        if not text:
+            return
+        link = self.repo.ui.config('tortoisehg', 'issue.link')
+        if link:
+            groups = self.issuedict.get(text, [text])
+            link, num = re.subn(r'\{(\d+)\}', lambda m:
+                groups[int(m.group(1))], link)
+            if not num:
+                link += text
+            shlib.browse_url(link)
+
+    def csetlink_event(self, tag, widget, event, liter):
+        if event.type != gtk.gdk.BUTTON_RELEASE:
+            return
+        text = self.get_link_text(tag, widget, liter)
+        if not text:
+            return
+        try:
+            rev = self.repo[text].rev()
+            if self.graphview:
+                self.graphview.set_revision_id(rev, load=True)
+            else:
+                self.load_details(rev)
+        except error.RepoError:
+            pass
+
+    def get_link_text(self, tag, widget, liter):
+        text_buffer = widget.get_buffer()
+        beg = liter.copy()
+        while not beg.begins_tag(tag):
+            beg.backward_char()
+        end = liter.copy()
+        while not end.ends_tag(tag):
+            end.forward_char()
+        text = text_buffer.get_text(beg, end)
+        return text
+
     def file_button_release(self, widget, event):
         if event.button == 3 and not (event.state & (gtk.gdk.SHIFT_MASK |
             gtk.gdk.CONTROL_MASK)):
@@ -730,10 +891,10 @@ class ChangeSet(gdialog.GDialog):
         try:
             fctx = ctx.filectx(self.curfile)
             has_filelog = fctx.filelog().linkrev(fctx.filerev()) == ctx.rev()
-        except hglib.LookupError:
+        except error.LookupError:
             has_filelog = False
         self.ann_menu.set_sensitive(has_filelog)
-        self.save_menu.set_sensitive(has_filelog)
+        self.msave.set_sensitive(has_filelog)
         return True
 
     def thgdiff(self, treeview):
@@ -762,10 +923,14 @@ class ChangeSet(gdialog.GDialog):
         result = gtklib.NativeSaveFileDialogWrapper(title=_("Save file to"),
                                                     initial=self.cwd,
                                                     filename=filename).run()
-        if result:
+        if not result:
+            return
+        try:
             q = Queue.Queue()
             hglib.hgcmd_toq(q, 'cat', '--rev',
                 str(self.currev), '--output', hglib.fromutf(result), self.curfile)
+        except (util.Abort, IOError), e:
+            gdialog.Prompt(_('Unable to save file'), str(e), self).run()
 
     def diff_to_local(self, menuitem):
         if not self.curfile:
@@ -818,12 +983,16 @@ class ChangeSet(gdialog.GDialog):
         if self.glog_parent:
             # If this changeset browser is embedded in glog, send
             # send this event to the main app
-            opts = {'pats' : [self.curfile]}
-            self.glog_parent.filtercombo.set_active(1)
-            self.glog_parent.filterentry.set_text(self.curfile)
-            self.glog_parent.custombutton.set_active(True)
-            self.glog_parent.filter = 'custom'
-            self.glog_parent.reload_log(**opts)
+            path = hglib.escapepath(self.curfile)
+            fname = hglib.toutf(path)
+            opts = {'pats': [fname]}
+            explorer = self.glog_parent
+            explorer.filter = 'custom'
+            explorer.filtercombo.set_active(1)
+            explorer.filterentry.set_text(fname)
+            explorer.filterbar.get_button('custom').set_active(True)
+            explorer.filter_entry_activated(explorer.filterentry,
+                                            explorer.filtercombo)
         else:
             # Else launch our own glog instance
             from tortoisehg.hgtk import history
@@ -845,21 +1014,14 @@ class ChangeSet(gdialog.GDialog):
         shlib.shell_notify([self.repo.wjoin(self.curfile)])
 
     def add_to_popup(self, textview, menu):
-        menu_items = (('----', None),
-                      (_('Toggle _Wordwrap'), self.toggle_wordwrap),
-                     )
-        for label, handler in menu_items:
-            if label == '----':
-                menuitem = gtk.SeparatorMenuItem()
-            else:
-                menuitem = gtk.MenuItem(label)
-            if handler:
-                menuitem.connect('activate', handler)
-            menu.append(menuitem)
+        menu.append(gtk.SeparatorMenuItem())
+        check = self.textview.get_wrap_mode() == gtk.WRAP_WORD
+        menu.append(gtklib.create_menuitem(_('Enable _Wordwrap'),
+                                           self.wordwrap_activated,
+                                           ascheck=True, check=check))
         menu.show_all()
 
-    def toggle_wordwrap(self, sender):
-        if self.textview.get_wrap_mode() != gtk.WRAP_NONE:
-            self.textview.set_wrap_mode(gtk.WRAP_NONE)
-        else:
-            self.textview.set_wrap_mode(gtk.WRAP_WORD)
+    def wordwrap_activated(self, menuitem):
+        mode = self.textview.get_wrap_mode() != gtk.WRAP_WORD \
+                    and gtk.WRAP_WORD or gtk.WRAP_NONE
+        self.textview.set_wrap_mode(mode)
