@@ -12,15 +12,15 @@ import os
 import gtk
 import binascii
 
-from mercurial import patch, util
-from mercurial.node import short, hex
+from mercurial import patch, util, error
+from mercurial.node import hex
 
 from tortoisehg.util.i18n import _
 from tortoisehg.util import hglib, paths
 
 from tortoisehg.hgtk import gtklib
 
-PANEL_DEFAULT = ('rev', 'summary', 'user', 'dateage', 'branch', 'tags', 'transplant')
+PANEL_DEFAULT = ('rev', 'summary', 'user', 'dateage', 'branch', 'tags', 'transplant', 'p4', 'svn')
 
 def create(repo, target=None, style=None, custom=None, **kargs):
     return Factory(repo, custom, style, target, **kargs)()
@@ -113,7 +113,7 @@ def ChangesetContext(repo, rev):
         return None
     try:
         ctx = repo[rev]
-    except (hglib.LookupError, hglib.RepoLookupError, hglib.RepoError):
+    except (error.LookupError, error.RepoLookupError, error.RepoError):
         ctx = None
     return ctx
 
@@ -165,13 +165,13 @@ class patchctx(object):
                 continue
             try:
                 self._parents.append(repo[p])
-            except (hglib.LookupError, hglib.RepoLookupError, hglib.RepoError):
+            except (error.LookupError, error.RepoLookupError, error.RepoError):
                 self._parents.append(p)
 
     def __str__(self):
         node = self.node()
         if node:
-            return short(node)
+            return node[:12]
         return ''
 
     def __int__(self):
@@ -191,6 +191,7 @@ class patchctx(object):
     def tags(self): return ()
     def parents(self): return self._parents
     def children(self): return ()
+    def extra(self): return {}
 
 class SummaryInfo(object):
 
@@ -199,7 +200,9 @@ class SummaryInfo(object):
               'user': _('User:'), 'date': _('Date:'),'age': _('Age:'),
               'dateage': _('Date:'), 'branch': _('Branch:'),
               'tags': _('Tags:'), 'rawbranch': _('Branch:'),
-              'rawtags': _('Tags:'), 'transplant': _('Transplant:')}
+              'rawtags': _('Tags:'), 'transplant': _('Transplant:'),
+              'p4': _('Perforce:'), 'svn': _('Subversion:'),
+              'shortuser': _('User:')}
 
     def __init__(self):
         pass
@@ -212,7 +215,9 @@ class SummaryInfo(object):
             if item == 'rev':
                 revnum = self.get_data('revnum', *args)
                 revid = self.get_data('revid', *args)
-                return (revnum, revid)
+                if revid:
+                    return (revnum, revid)
+                return None
             elif item == 'revnum':
                 return ctx.rev()
             elif item == 'revid':
@@ -226,6 +231,8 @@ class SummaryInfo(object):
                 return hglib.toutf(hglib.tounicode(value)[:80])
             elif item == 'user':
                 return hglib.toutf(ctx.user())
+            elif item == 'shortuser':
+                return hglib.toutf(hglib.username(ctx.user()))
             elif item == 'dateage':
                 date = self.get_data('date', *args)
                 age = self.get_data('age', *args)
@@ -243,18 +250,21 @@ class SummaryInfo(object):
                     return hglib.age(date)
                 return None
             elif item == 'rawbranch':
-                return hglib.toutf(ctx.branch())
+                value = ctx.branch()
+                if value:
+                    return hglib.toutf(value)
+                return None
             elif item == 'branch':
                 value = self.get_data('rawbranch', *args)
                 if value:
                     repo = ctx._repo
                     if ctx.node() not in repo.branchtags().values():
                         return None
-                    dblist = self.get_config(repo, 'tortoisehg', 'deadbranch')
-                    if dblist and value in [hglib.toutf(b.strip()) \
-                                            for b in dblist.split(',')]:
+                    dblist = hglib.getdeadbranch(repo.ui)
+                    if value in dblist:
                         return None
-                return value
+                    return value
+                return None
             elif item == 'rawtags':
                 value = [hglib.toutf(tag) for tag in ctx.tags()]
                 if len(value) == 0:
@@ -263,13 +273,12 @@ class SummaryInfo(object):
             elif item == 'tags':
                 value = self.get_data('rawtags', *args)
                 if value:
-                    repo = ctx._repo
-                    htags = self.get_config(repo, 'tortoisehg', 'hidetags', '')
-                    htags = [hglib.toutf(b.strip()) for b in htags.split()]
-                    value = [tag for tag in value if tag not in htags]
-                    if len(value) == 0:
+                    htlist = hglib.gethidetags(ctx._repo.ui)
+                    tags = [tag for tag in value if tag not in htlist]
+                    if len(tags) == 0:
                         return None
-                return value
+                    return tags
+                return None
             elif item == 'transplant':
                 extra = ctx.extra()
                 try:
@@ -279,8 +288,20 @@ class SummaryInfo(object):
                 except KeyError:
                     pass
                 return None
+            elif item == 'p4':
+                extra = ctx.extra()
+                p4cl = extra.get('p4', None)
+                return p4cl and ('changelist %s' % p4cl)
+            elif item == 'svn':
+                extra = ctx.extra()
+                cvt = extra.get('convert_revision', '')
+                if cvt.startswith('svn:'):
+                    return cvt.split('/', 1)[-1]
+                else:
+                    return None
             elif item == 'ishead':
-                return len(ctx.children()) == 0
+                childbranches = [cctx.branch() for cctx in ctx.children()]
+                return ctx.branch() not in childbranches
             raise UnknownItem(item)
         if custom.has_key('data') and not kargs.get('usepreset', False):
             try:
@@ -321,19 +342,22 @@ class SummaryInfo(object):
             if item == 'rev':
                 revnum, revid = value
                 revid = gtklib.markup(revid, **mono)
-                return '%s (%s)' % (revnum, revid)
+                if revnum is not None and revid is not None:
+                    return '%s (%s)' % (revnum, revid)
+                return '%s' % revid
             elif item in ('revid', 'transplant'):
                 return gtklib.markup(value, **mono)
-            elif item == 'revnum':
+            elif item in ('revnum', 'p4', 'svn'):
                 return str(value)
             elif item in ('rawbranch', 'branch'):
                 return gtklib.markup(' %s ' % value, color='black',
-                                     background='#aaffaa')
+                                     background=gtklib.PGREEN)
             elif item in ('rawtags', 'tags'):
-                opts = dict(color='black', background='#ffffaa')
+                opts = dict(color='black', background=gtklib.PYELLOW)
                 tags = [gtklib.markup(' %s ' % tag, **opts) for tag in value]
                 return ' '.join(tags)
-            elif item in ('desc', 'summary', 'user', 'date', 'age'):
+            elif item in ('desc', 'summary', 'user', 'shortuser',
+                          'date', 'age'):
                 return gtklib.markup(value)
             elif item == 'dateage':
                 return gtklib.markup('%s (%s)' % value)
@@ -418,19 +442,8 @@ class CachedSummaryInfo(SummaryInfo):
     def get_widget(self, *args, **kargs):
         return self.try_cache('widget', SummaryInfo.get_widget, *args, **kargs)
 
-    def get_config(self, repo, section, key, default=None, untrusted=False):
-        cachekey = repo.root + section + key
-        try:
-            return self.confcache[cachekey]
-        except KeyError:
-            pass
-        value = repo.ui.config(section, key, default, untrusted)
-        self.confcache[cachekey] = value
-        return value
-
     def clear_cache(self):
         self.cache = {}
-        self.confcache = {}
 
 class SummaryBase(object):
 
