@@ -1,4 +1,4 @@
-# gdialog.py - base dialog for gtools
+# gdialog.py - base window & dialogs for gtools
 #
 # Copyright 2007 Brad Schick, brad at gmail . com
 # Copyright 2008 Steve Borho <steve@borho.org>
@@ -14,11 +14,12 @@ import shutil
 import tempfile
 import gtk
 import atexit
+import pango
 
-from mercurial import cmdutil, util, ui, hg, commands
+from mercurial import cmdutil, util, ui, hg, commands, error
 
 from tortoisehg.util.i18n import _
-from tortoisehg.util import settings, hglib, paths
+from tortoisehg.util import settings, hglib, paths, shlib
 
 from tortoisehg.hgtk import gtklib
 
@@ -28,7 +29,6 @@ class SimpleMessage(gtklib.MessageDialog):
         self.destroy()
         return response
 
-
 class Prompt(SimpleMessage):
     def __init__(self, title, message, parent, type=gtk.MESSAGE_INFO):
         SimpleMessage.__init__(self, parent, gtk.DIALOG_MODAL,
@@ -40,9 +40,12 @@ class Prompt(SimpleMessage):
         key, modifier = gtk.accelerator_parse(mod+'Return')
         accel_group = gtk.AccelGroup()
         self.add_accel_group(accel_group)
-        buttons = self.get_children()[0].get_children()[1].get_children()
-        buttons[0].add_accelerator('clicked', accel_group, key,
-                modifier, gtk.ACCEL_VISIBLE)
+        try:
+            buttons = self.get_children()[0].get_children()[1].get_children()
+            buttons[0].add_accelerator('clicked', accel_group, key,
+                                       modifier, gtk.ACCEL_VISIBLE)
+        except IndexError:
+            pass
 
 class CustomPrompt(gtk.MessageDialog):
     ''' Custom prompt dialog.  Provide a list of choices with ampersands
@@ -51,10 +54,19 @@ class CustomPrompt(gtk.MessageDialog):
     '''
     # ret = CustomPrompt('Title', 'Message', self, ('&Yes', 'N&o'), 1).run()
     # ret will be (gtk.RESPONSE_DELETE_EVENT, 0 (for yes), or 1 (for no)
-    def __init__(self, title, message, parent, choices, default=None, esc=None):
+    def __init__(self, title, message, parent, choices, default=None,
+                 esc=None, files=None):
         gtk.MessageDialog.__init__(self, parent, gtk.DIALOG_MODAL,
                 gtk.MESSAGE_QUESTION)
         self.set_title(hglib.toutf(title))
+        if files:
+            msg = ''
+            for i, file in enumerate(files):
+                msg += '   %s\n' % file
+                if i == 9:
+                    msg += '   ...\n'
+                    break
+            message += '\n\n' + msg
         self.format_secondary_markup(gtklib.markup(hglib.toutf(message),
                                      weight='bold'))
         accel_group = gtk.AccelGroup()
@@ -96,14 +108,14 @@ class Confirm(SimpleMessage):
         accel_group = gtk.AccelGroup()
         self.add_accel_group(accel_group)
         buttons = self.get_children()[0].get_children()[1].get_children()
-        buttons[1].add_accelerator("clicked", accel_group, ord("y"),
+        buttons[1].add_accelerator('clicked', accel_group, ord('y'),
                               0, gtk.ACCEL_VISIBLE)
-        buttons[0].add_accelerator("clicked", accel_group, ord("n"),
+        buttons[0].add_accelerator('clicked', accel_group, ord('n'),
                               0, gtk.ACCEL_VISIBLE)
 
-
-class GDialog(gtk.Window):
-    """GTK+ based dialog for displaying mercurial information
+class GWindow(gtk.Window):
+    """
+    gtk.Window based window for displaying mercurial information
 
     The following methods are meant to be overridden by subclasses. At this
     point GCommit is really the only intended subclass.
@@ -255,21 +267,14 @@ class GDialog(gtk.Window):
         return self.opts.get(opt, False)
 
     def _parse_config(self):
-        # defaults
-        self.fontcomment = 'monospace 10'
-        self.fontdiff = 'monospace 10'
-        self.fontlist = 'Sans 9'
-        self.diffbottom = ''
-
-        for attr, setting in self.ui.configitems('gtools'):
-            if setting : setattr(self, attr, setting)
-
-        if not self.diffbottom:
+        self.rawfonts = hglib.getfontconfig(self.ui)
+        self.fonts = {}
+        for name, val in self.rawfonts.items():
+            self.fonts[name[4:]] = pango.FontDescription(val)
+        try:
+            self.diffbottom = self.ui.configbool('gtools', 'diffbottom', False)
+        except error.ConfigError:
             self.diffbottom = False
-        elif self.diffbottom.lower() == 'false' or self.diffbottom == '0':
-            self.diffbottom = False
-        else:
-            self.diffbottom = True
 
 
     def _parse_opts(self):
@@ -323,11 +328,8 @@ class GDialog(gtk.Window):
         if tip:
             tbutton.set_tooltip(self.tooltips, tip)
         if icon:
-            path = paths.get_tortoise_icon(icon)
-            if path:
-                image = gtk.Image()
-                image.set_from_file(path)
-                tbutton.set_icon_widget(image)
+            image = gtklib.get_icon_image(icon)
+            tbutton.set_icon_widget(image)
         tbutton.set_use_underline(True)
         tbutton.set_label(label)
         tbutton.connect('clicked', handler, userdata)
@@ -380,20 +382,28 @@ class GDialog(gtk.Window):
     def get_reponame(self):
         return hglib.get_reponame(self.repo)
 
-    def helpcontents(self, item):
+    def helpindex(self, item):
+        self.helpcontents(item, 'index.html')
+
+    def helpcontents(self, item, url=None):
         'User selected Help->Contents from menu bar'
-        url = self.get_help_url()
         if not url:
-            return
+            url = self.get_help_url()
+            if not url:
+                return
         if not url.startswith('http'):
-            url = 'http://tortoisehg.org/manual/0.9/' + url
-        from tortoisehg.hgtk import about
-        about.browse_url(url)
+            fullurl = 'http://tortoisehg.org/manual/1.0/' + url
+            # Use local CHM file if it can be found
+            if os.name == 'nt' and paths.bin_path:
+                chm = os.path.join(paths.bin_path, 'doc', 'TortoiseHg.chm')
+                if os.path.exists(chm):
+                    fullurl = (r'mk:@MSITStore:%s::/' % chm) + url
+        shlib.browse_url(fullurl)
 
     def launch(self, item, app):
         import sys
         # Spawn background process and exit
-        if hasattr(sys, "frozen"):
+        if hasattr(sys, 'frozen'):
             args = [sys.argv[0], app]
         else:
             args = [sys.executable] + [sys.argv[0], app]
@@ -456,6 +466,8 @@ class GDialog(gtk.Window):
         tbuttons =  self.get_tbbuttons()
         for tbutton in tbuttons:
             toolbar.insert(tbutton, -1)
+        for x in toolbar.get_children():
+            x.set_homogeneous(True)
         self.toolbar = toolbar
 
         # Subclass returns the main body
@@ -467,78 +479,56 @@ class GDialog(gtk.Window):
         menus = self.get_menu_list()
         if menus:
             allmenus = [
-          (_('_Tools'),
-           [dict(text=_('Repository Explorer'), func=self.launch, args=['log'],
-                icon='menulog.ico'),
-            dict(text=_('Commit'), func=self.launch, args=['commit'],
-                icon='menucommit.ico'),
-            dict(text=_('Datamine'), func=self.launch, args=['datamine'],
-                icon='menurepobrowse.ico'),
-            dict(text=_('Recovery'), func=self.launch, args=['recover'],
-                icon='general.ico'),
-            dict(text=_('Serve'), func=self.launch, args=['serve'],
-                icon='proxy.ico'),
-            dict(text=_('Shelve'), func=self.launch, args=['shelve'],
-                icon='shelve.ico'),
-            dict(text=_('Synchronize'), func=self.launch, args=['synch'],
-                icon='menusynch.ico'),
-            dict(text=_('Settings'), func=self.launch, args=['repoconfig'],
-                icon='settings_repo.ico')])
-           ] + menus + [
-          (_('_Help'),
-           [dict(text=_('Contents'), func=self.helpcontents,
-                icon=gtk.STOCK_INFO),
-            dict(text=_('About'), func=self.launch, args=['about'],
-                icon=gtk.STOCK_ABOUT)])
-          ]
-            menubar = gtk.MenuBar()
-            for title, items in allmenus:
-                m_items = gtklib.MenuItems()
-                for d in items:
+            dict(text=_('_Tools'), subitems=[
+                dict(text=_('Repository Explorer'), func=self.launch,
+                    args=['log'], icon='menulog.ico'),
+                dict(text=_('Commit'), func=self.launch,
+                    args=['commit'], icon='menucommit.ico'),
+                dict(text=_('Datamine'), func=self.launch,
+                    args=['datamine'], icon='menurepobrowse.ico'),
+                dict(text=_('Recovery'), func=self.launch,
+                    args=['recover'], icon='general.ico'),
+                dict(text=_('Serve'), func=self.launch,
+                    args=['serve'], icon='proxy.ico'),
+                dict(text=_('Shelve'), func=self.launch,
+                    args=['shelve'], icon='shelve.ico'),
+                dict(text=_('Synchronize'), func=self.launch,
+                    args=['synch'], icon='menusynch.ico'),
+                dict(text=_('Settings'), func=self.launch,
+                    args=['repoconfig'], icon='settings_repo.ico')])
+                ] + menus + [
+            dict(text=_('_Help'), subitems=[
+                dict(text=_('Contents'), func=self.helpcontents,
+                    icon=gtk.STOCK_INFO),
+                dict(text=_('Index'), func=self.helpindex,
+                    icon=gtk.STOCK_HELP),
+                dict(text=_('About'), func=self.launch,
+                    args=['about'], icon=gtk.STOCK_ABOUT)])
+                ]
+            def build(menus, nobuild=False):
+                mb = gtklib.MenuBuilder()
+                for d in menus:
                     text = d['text']
-                    name = d.get('name')
-                    func = d.get('func')
-                    ascheck = d.get('ascheck', False)
-                    args = d.get('args', [])
-                    icon = d.get('icon')
-                    check = d.get('check', False)
-                    sensitive = d.get('sensitive', None)
                     if text == '----':
-                        item = gtk.SeparatorMenuItem()
+                        mb.append_sep()
+                        continue
+                    subitems = d.get('subitems')
+                    if subitems:
+                        sub = build(subitems)
+                        item = mb.append_submenu(text, sub, **d)
                     else:
-                        if ascheck:
-                            item = gtk.CheckMenuItem(text)
-                            item.set_active(check)
-                        elif icon:
-                            item = gtk.ImageMenuItem(text)
-                            if icon.startswith('gtk'):
-                                img = gtk.image_new_from_stock(
-                                    icon, gtk.ICON_SIZE_MENU)
-                            else:
-                                img = gtk.Image()
-                                ico = paths.get_tortoise_icon(icon)
-                                if ico:
-                                    try:
-                                        width, height = gtk.icon_size_lookup(
-                                            gtk.ICON_SIZE_MENU)
-                                        pixbuf = gtk.gdk.pixbuf_new_from_file_at_size(
-                                            ico, width, height)
-                                        img.set_from_pixbuf(pixbuf)
-                                    except:
-                                        # don't let broken gtk+ to break dialogs
-                                        pass
-                            item.set_image(img)
-                        else:
-                            item = gtk.MenuItem(text)
-                        if sensitive is not None:
-                            item.set_sensitive(sensitive)
-                        item.connect('activate', func, *args)
-                        if name:
-                            self.menuitems[name] = item
-                    m_items.append(item)
-                item = gtk.MenuItem(title)
-                item.set_submenu(m_items.create_menu())
-                menubar.append(item)
+                        if 'rg' in d:
+                            d.update(group=self.menuitems.get(d['rg']))
+                        item = mb.append(text, d.get('func'), **d)
+                    name = d.get('name')
+                    if name:
+                        self.menuitems[name] = item
+                if nobuild:
+                    return mb.get_menus()
+                return mb.build()
+            menubar = gtk.MenuBar()
+            for menu in build(allmenus, nobuild=True):
+                menubar.append(menu)
 
         vbox = gtk.VBox(False, 0)
         self.add(vbox)
@@ -635,8 +625,8 @@ class GDialog(gtk.Window):
             ctx = repo[node]
             mf = ctx.manifest()
             dirname = os.path.basename(repo.root)
-            if dirname == "":
-                dirname = "root"
+            if dirname == '':
+                dirname = 'root'
             dirname = '%s.%s' % (dirname, str(ctx))
             base = os.path.join(tmproot, dirname)
             try:
@@ -674,10 +664,9 @@ class GDialog(gtk.Window):
                         copynode, self.tmproot)
                 pathroot = os.path.join(self.tmproot, copydir)
 
-            paths = ['"'+os.path.join(pathroot, f)+'"' for f in files]
-            command = editor + ' ' + ' '.join(paths)
+            paths = ['"%s"' % os.path.join(pathroot, f) for f in files]
+            command = '%s %s' % (editor, ' '.join(paths))
             util.system(command,
-                        environ={'HGUSER': self.ui.username()},
                         onerr=self.ui, errprefix=_('edit failed'))
 
         editor = (self.ui.config('tortoisehg', 'editor') or
@@ -702,3 +691,215 @@ class GDialog(gtk.Window):
         thread.setDaemon(True)
         thread.start()
 
+MODE_NORMAL  = 'normal'
+MODE_WORKING = 'working'
+
+RESPONSE_FORCE_CLOSE = 1024
+
+class GDialog(gtk.Dialog):
+    """ gtk.Dialog based window for executing mercurial operations """
+    def __init__(self, resizable=False, norepo=False):
+        gtk.Dialog.__init__(self)
+        gtklib.set_tortoise_icon(self, self.get_icon())
+        gtklib.set_tortoise_keys(self)
+        self.set_resizable(resizable)
+        self.set_has_separator(False)
+        self.earlyout = False
+
+        self.ui = ui.ui()
+        if norepo:
+            repo = None
+        else:
+            try:
+                repo = hg.repository(self.ui, path=paths.find_root())
+            except error.RepoError:
+                gtklib.idle_add_single_call(self.destroy)
+                return
+        self.repo = repo
+        self.after_done = True
+
+        # persistent settings
+        name = self.get_setting_name()
+        if name:
+            self.settings = settings.Settings(name)
+
+        # dialog size
+        defsize = self.get_defsize()
+        if defsize != (-1, -1):
+            self.set_default_size(*defsize)
+
+        # signal handler
+        self.connect('realize', self.realized)
+
+    ### Overridable Functions ###
+
+    def get_title(self, reponame):
+        return 'TortoiseHg - %s' % reponame
+
+    def get_icon(self):
+        return 'thg_logo.ico'
+
+    def get_defsize(self):
+        return (-1, -1)
+
+    def get_setting_name(self):
+        return None
+
+    def get_body(self, vbox):
+        pass
+
+    def get_extras(self, vbox):
+        pass
+
+    def get_buttons(self):
+        return []
+
+    def get_default_button(self):
+        return None
+
+    def get_action_map(self):
+        return {}
+
+    def switch_to(self, normal, working, cmd):
+        pass
+
+    def command_done(self, returncode, useraborted, *args):
+        pass
+
+    def before_close(self):
+        return True
+
+    def load_settings(self):
+        pass
+
+    def store_settings(self):
+        pass
+
+    ### Public Functions ###
+
+    def set_notify_func(self, func, *args):
+        self.notify_func = func
+        self.notify_args = args
+
+    def set_after_done(self, close):
+        self.after_done = close
+
+    ### Internal Functions ###
+
+    def after_init(self):
+        from tortoisehg.hgtk import hgcmd
+        self.get_extras(self.vbox)
+
+        # CmdWidget
+        self.cmd = hgcmd.CmdWidget()
+        self.cmd.show_all()
+        self.cmd.hide()
+        self.vbox.pack_start(self.cmd, False, False, 6)
+
+        # add Abort button
+        self.action_area.add(self.buttons['abort'])
+
+    def do_switch_to(self, mode, cmd=True):
+        if mode == MODE_NORMAL:
+            normal = True
+        elif mode == MODE_WORKING:
+            normal = False
+            self.buttons['abort'].grab_focus()
+        else:
+            raise _('unknown mode name: %s') % mode
+        working = not normal
+
+        if cmd:
+            self.cmd.set_property('visible', working)
+        self.buttons['abort'].set_property('visible', working)
+
+        self.switch_to(normal, working, cmd)
+
+    def abort(self):
+        self.cmd.stop()
+        self.cmd.show_log()
+        self.do_switch_to(MODE_NORMAL, cmd=False)
+
+    def execute_command(self, cmdline, *args):
+        def cmd_done(returncode, useraborted):
+            self.do_switch_to(MODE_NORMAL, cmd=False)
+            self.command_done(returncode, useraborted, *args)
+            if hasattr(self, 'notify_func') and self.notify_func:
+                self.notify_func(*self.notify_args)
+            if self.after_done and returncode == 0:
+                if not self.cmd.is_show_log():
+                    self.response(gtk.RESPONSE_CLOSE)
+        self.do_switch_to(MODE_WORKING)
+        self.cmd.execute(cmdline, cmd_done)
+
+    ### Signal Handler ###
+
+    def realized(self, *args):
+        # set title
+        reponame = self.repo and hglib.get_reponame(self.repo) or ''
+        self.set_title(self.get_title(reponame))
+
+        # add user-defined buttons
+        self.buttons = {}
+        for name, label, res in self.get_buttons():
+            btn = self.add_button(label, res)
+            self.buttons[name] = btn
+
+        # create Abort button (add later)
+        btn = gtk.Button(_('Abort'))
+        btn.connect('clicked', lambda *a: self.response(gtk.RESPONSE_CANCEL))
+        self.buttons['abort'] = btn
+
+        # construct dialog body
+        self.get_body(self.vbox)
+        if self.earlyout:
+            gtklib.idle_add_single_call(self.destroy)
+            return
+
+        # focus on default button if needs
+        name = self.get_default_button()
+        if name:
+            btn = self.buttons.get(name)
+            if btn:
+                btn.grab_focus()
+
+        # signal handler
+        self.connect('response', self.dialog_response)
+
+        # prepare to show
+        self.load_settings()
+        self.vbox.show_all()
+        gtklib.idle_add_single_call(self.after_init)
+
+    def dialog_response(self, dialog, response_id):
+        # User-defined buttons
+        actmap = self.get_action_map()
+        if actmap.has_key(response_id):
+            actmap[response_id]()
+        # Forced close
+        elif response_id == RESPONSE_FORCE_CLOSE:
+            if self.cmd.is_alive():
+                self.abort()
+            self.store_settings()
+            self.destroy()
+            return # close dialog
+        # Cancel button or dialog closing by the user
+        elif response_id in (gtk.RESPONSE_CLOSE, gtk.RESPONSE_DELETE_EVENT):
+            if self.cmd.is_alive():
+                ret = Confirm(_('Confirm Abort'), [], self,
+                              _('Do you want to abort?')).run()
+                if ret == gtk.RESPONSE_YES:
+                    self.abort()
+            else:
+                close = self.before_close()
+                if close:
+                    self.store_settings()
+                    self.destroy()
+                    return # close dialog
+        # Abort button
+        elif response_id == gtk.RESPONSE_CANCEL:
+            self.abort()
+        else:
+            raise _('unexpected response id: %s') % response_id
+
+        self.run() # don't close dialog
