@@ -27,6 +27,7 @@ from tortoisehg.hgtk.logview import treemodel
 from tortoisehg.hgtk.logview.graphcell import CellRendererGraph
 from tortoisehg.hgtk.logview.revgraph import *
 
+COLS = 'graph rev id revhex branch changes msg user date utc age tag'
 
 class TreeView(gtk.ScrolledWindow):
 
@@ -77,15 +78,25 @@ class TreeView(gtk.ScrolledWindow):
                                  False,
                                  gobject.PARAM_READWRITE),
         'id-column-visible': (gobject.TYPE_BOOLEAN,
-                                 'Tags',
+                                 'ID',
                                  'Show revision ID column',
                                  False,
                                  gobject.PARAM_READWRITE),
+        'revhex-column-visible': (gobject.TYPE_BOOLEAN,
+                                 'Rev/ID',
+                                 'Show revision number/ID column',
+                                 False,
+                                 gobject.PARAM_READWRITE),
         'branch-column-visible': (gobject.TYPE_BOOLEAN,
                                  'Branch',
                                  'Show branch',
                                  False,
                                  gobject.PARAM_READWRITE),
+        'changes-column-visible': (gobject.TYPE_BOOLEAN,
+                                 'Changes',
+                                 'Show changes column',
+                                 False,
+                                 gobject.PARAM_READWRITE),
         'tag-column-visible': (gobject.TYPE_BOOLEAN,
                                  'Tags',
                                  'Show tag column',
@@ -110,7 +121,7 @@ class TreeView(gtk.ScrolledWindow):
                               ())
     }
 
-    def __init__(self, repo, limit=500, pbar=None):
+    def __init__(self, repo, limit=500, stbar=None):
         """Create a new TreeView.
 
         :param repo:  Repository object to show
@@ -123,7 +134,7 @@ class TreeView(gtk.ScrolledWindow):
         self.batchsize = limit
         self.repo = repo
         self.currevid = None
-        self.pbar = pbar
+        self.stbar = stbar
         self.grapher = None
         self.graphdata = []
         self.index = {}
@@ -131,9 +142,9 @@ class TreeView(gtk.ScrolledWindow):
                       'branch-color':False, 'show-graph':True }
         self.construct_treeview()
 
-    def set_repo(self, repo, pbar=None):
+    def set_repo(self, repo, stbar=None):
         self.repo = repo
-        self.pbar = pbar
+        self.stbar = stbar
 
     def search_in_tree(self, model, column, key, iter, data):
         """Searches all fields shown in the tree when the user hits crtr+f,
@@ -230,8 +241,8 @@ class TreeView(gtk.ScrolledWindow):
 
         if not len(self.graphdata):
             self.treeview.set_model(None)
-            if self.pbar is not None:
-                self.pbar.end()
+            if self.stbar is not None:
+                self.stbar.end()
             self.emit('revisions-loaded')
             return False
 
@@ -253,12 +264,12 @@ class TreeView(gtk.ScrolledWindow):
             self.emit('revisions-loaded')
         if revid is not None:
             self.set_revision_id(revid)
-        if self.pbar is not None:
-            self.pbar.end()
+        if self.stbar is not None:
+            self.stbar.end()
             revision_text = _('%(count)d of %(total)d Revisions') % {
                     'count': len(self.model),
                     'total': len(self.repo) }
-            self.pbar.set_right1_text(revision_text)
+            self.stbar.set_text(revision_text, name='rev')
         return False
 
     def do_get_property(self, property):
@@ -315,8 +326,8 @@ class TreeView(gtk.ScrolledWindow):
             return
         self.batchsize = size
         self.limit += self.batchsize
-        if self.pbar is not None:
-            self.pbar.begin()
+        if self.stbar is not None:
+            self.stbar.begin()
         gobject.idle_add(self.populate)
 
     def load_all_revisions(self):
@@ -324,8 +335,8 @@ class TreeView(gtk.ScrolledWindow):
             self.emit('revisions-loaded')
             return
         self.limit = None
-        if self.pbar is not None:
-            self.pbar.begin()
+        if self.stbar is not None:
+            self.stbar.begin()
         gobject.idle_add(self.populate)
 
     def scroll_to_revision(self, revid):
@@ -363,12 +374,12 @@ class TreeView(gtk.ScrolledWindow):
             hglib.invalidaterepo(self.repo)
             if len(self.repo) > 0:
                 self.create_log_generator(graphcol, pats, opts)
-                if self.pbar is not None:
-                    self.pbar.begin()
+                if self.stbar is not None:
+                    self.stbar.begin()
                 gobject.idle_add(self.populate, self.currevid)
             else:
                 self.treeview.set_model(None)
-                self.pbar.set_status_text(_('Repository is empty'))
+                self.stbar.set_text(_('Repository is empty'))
 
     def construct_treeview(self):
         self.treeview = gtk.TreeView()
@@ -423,6 +434,18 @@ class TreeView(gtk.ScrolledWindow):
         col.add_attribute(cell, "text", treemodel.HEXID)
         col.add_attribute(cell, "foreground", treemodel.FGCOLOR)
 
+        cell = gtk.CellRendererText()
+        cell.set_property("width-chars", 22)
+        cell.set_property("ellipsize", pango.ELLIPSIZE_END)
+        col = self.tvcolumns['revhex'] = gtk.TreeViewColumn(_('Rev/ID'))
+        col.set_visible(False)
+        col.set_resizable(True)
+        col.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
+        col.set_fixed_width(cell.get_size(self.treeview)[2])
+        col.pack_start(cell, expand=True)
+        col.add_attribute(cell, "foreground", treemodel.FGCOLOR)
+        col.add_attribute(cell, "markup", treemodel.REVHEX)
+
         cell = gtk.CellRendererText()
         cell.set_property("width-chars", 15)
         cell.set_property("ellipsize", pango.ELLIPSIZE_END)
@@ -436,7 +459,17 @@ class TreeView(gtk.ScrolledWindow):
         col.add_attribute(cell, "foreground", treemodel.FGCOLOR)
 
         cell = gtk.CellRendererText()
+        cell.set_property("width-chars", 10)
+        cell.set_property("ellipsize", pango.ELLIPSIZE_END)
+        col = self.tvcolumns['changes'] = gtk.TreeViewColumn(_('Changes'))
+        col.set_visible(False)
+        col.set_resizable(True)
+        col.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
+        col.set_fixed_width(cell.get_size(self.treeview)[2])
+        col.pack_start(cell, expand=True)
+        col.add_attribute(cell, "markup", treemodel.CHANGES)
 
+        cell = gtk.CellRendererText()
         cell.set_property("width-chars", 80)
         cell.set_property("ellipsize", pango.ELLIPSIZE_END)
         col = self.tvcolumns['msg'] = gtk.TreeViewColumn(_('Summary'))
@@ -483,7 +516,7 @@ class TreeView(gtk.ScrolledWindow):
         col.add_attribute(cell, "foreground", treemodel.FGCOLOR)
 
         cell = gtk.CellRendererText()
-        cell.set_property("width-chars", 10)
+        cell.set_property("width-chars", 16)
         cell.set_property("ellipsize", pango.ELLIPSIZE_END)
         col = self.tvcolumns['age'] = gtk.TreeViewColumn(_('Age'))
         col.set_visible(True)
@@ -506,23 +539,22 @@ class TreeView(gtk.ScrolledWindow):
         col.add_attribute(cell, "text", treemodel.TAGS)
         col.add_attribute(cell, "foreground", treemodel.FGCOLOR)
 
-        cols = 'graph rev id branch msg user date utc age tag'
-        self.columns = cols.split()
-
-        # append columns
-        for cn in self.columns:
-            c = self.tvcolumns[cn]
-            self.treeview.append_column(c)
+        self.columns = COLS.split()
 
     def set_columns(self, columns):
         if ' '.join(columns) != ' '.join(self.columns):
+            cols = self.treeview.get_columns()
             for cn in self.columns:
                 c = self.tvcolumns[cn]
-                self.treeview.remove_column(c)
+                if c in cols:
+                    self.treeview.remove_column(c)
             for cn in columns:
-                c = self.tvcolumns[cn]
-                self.treeview.append_column(c)
-                self.columns = columns
+                try:
+                    c = self.tvcolumns[cn]
+                    self.treeview.append_column(c)
+                except KeyError:
+                    continue
+            self.columns = columns
 
     def get_columns(self):
         return self.columns
