@@ -7,10 +7,11 @@
 
 import os
 import gtk
+import gtk.keysyms
 import gobject
 import pango
 
-from mercurial import extensions
+from mercurial import error
 
 from tortoisehg.util.i18n import _
 from tortoisehg.util import hglib
@@ -28,6 +29,15 @@ MQ_ESCAPED = 4
 INDEX_SEPARATOR = -1
 INDEX_QPARENT   = -2
 
+# Available operations for patch reordering
+MOVE_TOP    = 1
+MOVE_UP     = 2
+MOVE_DOWN   = 3
+MOVE_BOTTOM = 4
+
+# DnD target constans
+MQ_DND_URI_LIST = 1024
+
 class MQWidget(gtk.VBox):
 
     __gproperties__ = {
@@ -70,46 +80,42 @@ class MQWidget(gtk.VBox):
         'patch-selected': (gobject.SIGNAL_RUN_FIRST,
                            gobject.TYPE_NONE,
                            (int,  # revision number
-                            str)) # patch name
+                            str)), # patch name
+        'files-dropped': (gobject.SIGNAL_RUN_FIRST,
+                          gobject.TYPE_NONE,
+                          (object, # list of dropped files/dirs
+                           str))   # raw string data
     }
 
-    def __init__(self, repo, statusbar, accelgroup=None, tooltips=None):
+    def __init__(self, repo, accelgroup=None, tooltips=None):
         gtk.VBox.__init__(self)
 
         self.repo = repo
         self.mqloaded = hasattr(repo, 'mq')
-        self.statusbar = statusbar
-
-        try:
-            extensions.find('qup')
-            self.hasqup = True
-        except KeyError:
-            self.hasqup = False
 
         # top toolbar
         tbar = gtklib.SlimToolbar(tooltips)
 
         ## buttons
         self.btn = {}
-        popallbtn = tbar.append_stock(gtk.STOCK_GOTO_FIRST,
+        popallbtn = tbar.append_button(gtk.STOCK_GOTO_TOP,
                                       _('Unapply all patches'))
-        popallbtn.connect('clicked', self.popall_clicked)
+        popallbtn.connect('clicked', lambda *a: self.qpop(all=True))
         self.btn['popall'] = popallbtn
 
-        popbtn = tbar.append_stock(gtk.STOCK_GO_BACK,
+        popbtn = tbar.append_button(gtk.STOCK_GO_UP,
                                    _('Unapply last patch'))
-        popbtn.connect('clicked', self.pop_clicked)
+        popbtn.connect('clicked', lambda *a: self.qpop())
         self.btn['pop'] = popbtn
 
-        pushbtn = gtk.ToolButton(gtk.STOCK_GO_FORWARD)
-        pushbtn = tbar.append_stock(gtk.STOCK_GO_FORWARD,
+        pushbtn = tbar.append_button(gtk.STOCK_GO_DOWN,
                                     _('Apply next patch'))
-        pushbtn.connect('clicked', self.push_clicked)
+        pushbtn.connect('clicked', lambda *a: self.qpush())
         self.btn['push'] = pushbtn
 
-        pushallbtn = tbar.append_stock(gtk.STOCK_GOTO_LAST,
+        pushallbtn = tbar.append_button(gtk.STOCK_GOTO_BOTTOM,
                                        _('Apply all patches'))
-        pushallbtn.connect('clicked', self.pushall_clicked)
+        pushallbtn.connect('clicked', lambda *a: self.qpush(all=True))
         self.btn['pushall'] = pushallbtn
 
         ## separator
@@ -146,11 +152,21 @@ class MQWidget(gtk.VBox):
         # To support old PyGTK (<2.12)
         if hasattr(self.list, 'set_tooltip_column'):
             self.list.set_tooltip_column(MQ_ESCAPED)
+        self.list.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
+        if hasattr(self.list, 'set_rubber_banding'):
+            self.list.set_rubber_banding(True)
         self.list.connect('cursor-changed', self.list_sel_changed)
         self.list.connect('button-press-event', self.list_pressed)
+        self.list.connect('button-release-event', self.list_released)
         self.list.connect('row-activated', self.list_row_activated)
         self.list.connect('size-allocate', self.list_size_allocated)
 
+        ### dnd setup for patch list
+        targets = [('text/uri-list', 0, MQ_DND_URI_LIST)]
+        self.list.drag_dest_set(gtk.DEST_DEFAULT_MOTION | \
+             gtk.DEST_DEFAULT_DROP, targets, gtk.gdk.ACTION_MOVE)
+        self.list.connect('drag-data-received', self.dnd_received)
+
         self.cols = {}
         self.cells = {}
 
@@ -183,7 +199,7 @@ class MQWidget(gtk.VBox):
 
         addcol(_('#'), MQ_INDEX, right=True)
         addcol(_('st'), MQ_STATUS)
-        addcol(_('Name'), MQ_NAME, editfunc=cell_edited)
+        addcol(_('Patch'), MQ_NAME, editfunc=cell_edited)
         addcol(_('Summary'), MQ_SUMMARY, resizable=True)
 
         pane.add(self.list)
@@ -193,7 +209,7 @@ class MQWidget(gtk.VBox):
                                    tooltips=tooltips)
         mainbox.pack_start(self.cmd, False, False)
 
-        # accelerator
+        # accelerators
         if accelgroup:
             key, mod = gtk.accelerator_parse('F2')
             self.list.add_accelerator('thg-rename', accelgroup,
@@ -205,6 +221,18 @@ class MQWidget(gtk.VBox):
                     self.qrename_ui(model[paths[0]][MQ_NAME])
             self.list.connect('thg-rename', thgrename)
 
+            mod = gtk.gdk.CONTROL_MASK
+            def add(name, key, func, *args):
+                self.list.add_accelerator(name, accelgroup, key, mod, 0)
+                self.list.connect(name, lambda *a: func(*args))
+            add('mq-move-top', gtk.keysyms.Page_Up, self.qreorder_ui, MOVE_TOP)
+            add('mq-move-up', gtk.keysyms.Up, self.qreorder_ui, MOVE_UP)
+            add('mq-move-down', gtk.keysyms.Down, self.qreorder_ui, MOVE_DOWN)
+            add('mq-move-bottom', gtk.keysyms.Page_Down, self.qreorder_ui,
+                MOVE_BOTTOM)
+            add('mq-pop', gtk.keysyms.Left, self.qpop)
+            add('mq-push', gtk.keysyms.Right, self.qpush)
+
     ### public functions ###
 
     def refresh(self):
@@ -247,7 +275,8 @@ class MQWidget(gtk.VBox):
 
         # insert separator
         if top:
-            self.model.insert_after(top, (INDEX_SEPARATOR, None, None, None, None))
+            row = self.model.insert_after(top, (INDEX_SEPARATOR, None, None, None, None))
+            self.separator_pos = self.model.get_path(row)[0]
 
         # restore patch selection
         if selname:
@@ -258,21 +287,6 @@ class MQWidget(gtk.VBox):
         # update UI sensitives
         self.update_sensitives()
 
-        # report status
-        status_text = ''
-        idle_text = None
-        if self.has_mq():
-            nser = len(self.repo.mq.series)
-            if nser:
-                napp = len(self.repo.mq.applied)
-                status_text = _('%(count)d of %(total)d Patches applied') % {
-                        'count': napp, 'total': nser}
-                if napp:
-                    pn = self.get_qtip_patchname()
-                    idle_text = _("Patch '%s' applied") % pn
-        self.statusbar.set_right3_text(status_text)
-        self.statusbar.set_idle_text(idle_text)
-
     def set_repo(self, repo):
         self.repo = repo
 
@@ -317,22 +331,34 @@ class MQWidget(gtk.VBox):
         """
         [MQ] Execute 'qdelete' command.
 
-        patch: the patch name or an index to specify the patch.
+        patch: a patch name or an index number, or its list.
+        keep: if True, use '--keep' option. (default: False)
         """
-        if not self.has_patch():
+        if not patch or not self.has_patch():
+            return
+        if isinstance(patch, (basestring, int, long)):
+            patch = [patch]
+        unapplied = [p for p in patch if not self.is_applied(p)]
+        if not unapplied:
             return
         if not keep:
-            ret = gdialog.CustomPrompt(_('Confirm Delete'),
-                          _('Do you want to delete?'), None,
-                          (_('&Yes'), _('Yes (&keep)'),
-                          _('&Cancel')), default=2, esc=2).run()
+            buttons = (_('&Yes'), _('Yes (&keep)'), _('&Cancel'))
+            if len(unapplied) == 1:
+                data = dict(name=unapplied[0])
+                ret = gdialog.CustomPrompt(_('Confirm Delete'),
+                        _("Do you want to delete '%(name)s'?") % data,
+                        None, buttons, default=2, esc=2).run()
+            else:
+                ret = gdialog.CustomPrompt(_('Confirm Delete'),
+                        _('Do you want to delete these patches?'), None,
+                        buttons, default=2, esc=2, files=unapplied).run()
             if ret == 0:
                 pass
             elif ret == 1:
                 keep = True
             else:
                 return
-        cmdline = ['hg', 'qdelete', patch]
+        cmdline = ['hg', 'qdelete'] + unapplied
         if keep:
             cmdline.append('--keep')
         self.cmd.execute(cmdline, self.cmd_done, noemit=True)
@@ -401,51 +427,189 @@ class MQWidget(gtk.VBox):
         """
         [MQ] Execute 'qfold' command.
 
-        patch: the patch name or an index to specify the patch.
+        patch: a patch name or an index number, or its list.
         """
         if not patch or not self.has_applied():
             return
-        data = dict(target=patch, qtip=self.get_qtip_patchname())
-        ret = gdialog.Confirm(_('Confirm Fold'), [], None,
+        if isinstance(patch, (basestring, int, long)):
+            patch = [patch]
+        unapplied = [p for p in patch if not self.is_applied(p)]
+        if not unapplied:
+            return
+        data = dict(qtip=self.get_qtip_patchname())
+        if len(unapplied) == 1:
+            data.update(target=unapplied[0])
+            ret = gdialog.Confirm(_('Confirm Fold'), [], None,
                     _("Do you want to fold un-applied patch '%(target)s'"
                       " into current patch '%(qtip)s'?") % data).run()
+        else:
+            ret = gdialog.Confirm(_('Confirm Fold'), unapplied, None,
+                    _("Do you want to fold following un-applied patches"
+                      " into the current patch '%(qtip)s'?") % data).run()
         if ret != gtk.RESPONSE_YES:
             return
-        cmdline = ['hg', 'qfold', patch]
+        cmdline = ['hg', 'qfold'] + unapplied
         self.cmd.execute(cmdline, self.cmd_done)
 
-    def mknext(self, patch):
+    def qreorder(self, patch, op):
         """
-        [MQ] Execute 'qup patch'
+        [MQ] Reorder patches. This is NOT standard API of MQ.
 
-        patch: the patch name or an index to specify the patch.
+        patch: a patch name, or its list.
+        op: the operator for moving the patch: MOVE_TOP, MOVE_UP,
+            MOVE_DOWN or MOVE_BOTTOM.
         """
-        if not (self.hasqup and patch and self.is_operable()):
-            return
-        cmdline = ['hg', 'qup', patch]
-        self.cmd.execute(cmdline, self.cmd_done)
+        if isinstance(patch, (basestring)):
+            patch = [patch]
+        applied = [p for p in patch if self.is_applied(p)]
+        if not patch or not self.is_operable() or len(applied) > 0:
+            return False
+
+        # get min/max indices of selected patches in TreeView
+        model = self.model
+        lastidx = len(model) - 1
+        minidx, maxidx = lastidx, 0
+        for idx, row in enumerate(model):
+            if row[MQ_NAME] in patch:
+                if idx < minidx:
+                    minidx = idx
+                if maxidx < idx:
+                    maxidx = idx
+
+        # find index of first unapplied patch in TreeView
+        for i, row in enumerate(model):
+            if self.is_unapplied(row[MQ_NAME]):
+                uminidx = i
+                break
+        else:
+            return False
+
+        # check whether operation is possible
+        if (minidx == uminidx and op in (MOVE_TOP, MOVE_UP)) or \
+           (maxidx == lastidx and op in (MOVE_DOWN, MOVE_BOTTOM)):
+            return False
+
+        # determine dirty range
+        if op == MOVE_TOP:
+            dirty = range(uminidx, maxidx + 1)
+        elif op == MOVE_UP:
+            dirty = range(minidx - 1, maxidx + 1)
+        elif op == MOVE_DOWN:
+            dirty = range(minidx, maxidx + 2)
+        elif op == MOVE_BOTTOM:
+            dirty = range(minidx, lastidx + 1)
+        else:
+            raise _('invalid reorder operation: %s') % op
+
+        # prepare qdirty (dirty range in MQ series)
+        qdirty = range(model[dirty[0]][MQ_INDEX],
+                       model[dirty[-1]][MQ_INDEX] + 1)
+
+        # fill with safe indices (first-half)
+        nums = range(len(model))
+        neworder = nums[:dirty[0]]
+
+        # fill with new indices
+        def get_target_positions():
+            plist = []
+            for name in patch:
+                pos = self.get_pos_by_patchname(name)
+                plist.append(pos)
+            plist.sort()
+            return plist
+        if op in (MOVE_TOP, MOVE_UP):
+            # append target patch positions
+            positions = get_target_positions()
+            neworder += positions
+            # append remains
+            for pos in dirty:
+                if pos not in positions:
+                    neworder.append(pos)
+        else:
+            # prepare target patch positions
+            positions = get_target_positions()
+            # append non-target patch positions
+            for pos in dirty:
+                if pos not in positions:
+                    neworder.append(pos)
+            # append target patch positions
+            neworder += positions
+
+        # fill with safe indices (last-half)
+        neworder += nums[dirty[-1] + 1:]
+
+        # update TreeView
+        model.reorder(neworder)
+
+        # update series
+        q = self.repo.mq
+        series = q.full_series[:]
+        for pos, qpos in zip(dirty, qdirty):
+            q.full_series[qpos] = series[model[pos][MQ_INDEX]]
+        q.series_dirty = True
+        q.save_dirty()
+
+        # need to refresh
+        self.refresh()
+
+        # restore selection
+        sel = self.list.get_selection()
+        sel.unselect_all()
+        for row in model:
+            if row[MQ_NAME] in patch:
+                sel.select_path(row.path)
+
+        return True
+
+    def qreorder_ui(self, op):
+        """
+        [MQ] Reorder selected patches in the list.
+
+        Return True if succeed reordering; otherwise False.
+
+        op: the operator for moving the patch: MOVE_TOP, MOVE_UP,
+            MOVE_DOWN or MOVE_BOTTOM.
+        """
+        sel = self.list.get_selection()
+        if sel.count_selected_rows() == 0:
+            return False
+        patches = []
+        for path in sel.get_selected_rows()[1]:
+            row = self.model[path]
+            if row[MQ_INDEX] in (INDEX_SEPARATOR, INDEX_QPARENT):
+                continue
+            patches.append(row[MQ_NAME])
+        return self.qreorder(patches, op)
 
     def has_mq(self):
         return self.mqloaded and os.path.isdir(self.repo.mq.path)
 
     def has_patch(self):
-        """ return True if MQ has applicable patch """
-        if self.mqloaded:
-            return len(self.repo.mq.series) > 0
-        return False
+        """ return True if MQ has applicable patches """
+        return bool(self.get_num_patches())
 
     def has_applied(self):
         """ return True if MQ has applied patches """
-        if self.mqloaded:
-            return len(self.repo.mq.applied) > 0
-        return False
+        return bool(self.get_num_applied())
 
-    def number_applied(self):
+    def get_num_patches(self):
+        """ return the number of patches in patch queue """
+        if self.mqloaded:
+            return len(self.repo.mq.series)
+        return 0
+
+    def get_num_applied(self):
         """ return the number of applied patches """
         if self.mqloaded:
             return len(self.repo.mq.applied)
         return 0
 
+    def get_num_unapplied(self):
+        """ return the number of unapplied patches """
+        if self.mqloaded:
+            return self.get_num_patches() - self.get_num_applied()
+        return 0
+
     def is_operable(self):
         """ return True if MQ is operable """
         if self.mqloaded:
@@ -455,9 +619,20 @@ class MQWidget(gtk.VBox):
             return len(repo.mq.series) > 0
         return False
 
-    def is_qtip(self, patchname):
-        if patchname:
-            return patchname == self.get_qtip_patchname()
+    def is_applied(self, name):
+        if self.mqloaded:
+            return self.repo.mq.isapplied(name) is not None
+        return False
+
+    def is_unapplied(self, name):
+        if self.mqloaded:
+            return self.repo.mq.find_series(name) is not None and \
+                   not self.repo.mq.isapplied(name)
+        return False
+
+    def is_qtip(self, name):
+        if name:
+            return name == self.get_qtip_patchname()
         return False
 
     ### internal functions ###
@@ -472,10 +647,29 @@ class MQWidget(gtk.VBox):
 
     def get_path_by_patchname(self, name):
         """ return path has specified patch name """
-        return self.model.get_path(self.get_iter_by_patchname(name))
+        iter = self.get_iter_by_patchname(name)
+        if iter:
+            return self.model.get_path(iter)
+        return None
+
+    def get_row_by_patchname(self, name):
+        """ return row has specified patch name """
+        path = self.get_path_by_patchname(name)
+        if path:
+            return self.model[path]
+        return None
+
+    def get_pos_by_patchname(self, name):
+        """ return position of specified patch name """
+        if name:
+            for i, row in enumerate(self.model):
+                if row[MQ_NAME] == name:
+                    return i
+        return None
 
     def get_qtip_patchname(self):
-        if self.mqloaded and 'qtip' in self.repo.tags():
+        if self.mqloaded and self.get_num_applied() > 0 \
+                         and 'qtip' in self.repo.tags():
             return self.repo.mq.applied[-1].name
         return None
 
@@ -542,18 +736,26 @@ class MQWidget(gtk.VBox):
     def row_sep_func(self, model, iter, data=None):
         return model[iter][MQ_INDEX] == INDEX_SEPARATOR
 
-    def show_patch_cmenu(self, list, path):
+    def create_reorder_menu(self):
+        sub = gtklib.MenuBuilder()
+        sub.append(_('Top'), lambda *a: self.qreorder_ui(MOVE_TOP),
+                   gtk.STOCK_GOTO_TOP)
+        sub.append(_('Up'), lambda *a: self.qreorder_ui(MOVE_UP),
+                   gtk.STOCK_GO_UP)
+        sub.append(_('Down'), lambda *a: self.qreorder_ui(MOVE_DOWN),
+                   gtk.STOCK_GO_DOWN)
+        sub.append(_('Bottom'), lambda *a: self.qreorder_ui(MOVE_BOTTOM),
+                   gtk.STOCK_GOTO_BOTTOM)
+        return sub.build()
+
+    def show_patch_cmenu(self, path):
         row = self.model[path]
         if row[MQ_INDEX] == INDEX_SEPARATOR:
             return
 
-        menu = gtk.Menu()
-        def append(label, handler=None):
-            item = gtk.MenuItem(label, True)
-            item.set_border_width(1)
-            if handler:
-                item.connect('activate', handler, row)
-            menu.append(item)
+        m = gtklib.MenuBuilder()
+        def append(*args):
+            m.append(*args, **dict(args=[row]))
 
         is_operable = self.is_operable()
         has_patch = self.has_patch()
@@ -561,70 +763,87 @@ class MQWidget(gtk.VBox):
         is_qtip = self.is_qtip(row[MQ_NAME])
         is_qparent = row[MQ_INDEX] == INDEX_QPARENT
         is_applied = row[MQ_STATUS] == 'A'
-        is_next = row[MQ_INDEX] == self.number_applied()
 
         if is_operable and not is_qtip and (not is_qparent or has_applied):
-            append(_('_goto'), self.goto_activated)
+            append(_('_Goto'), self.goto_activated, gtk.STOCK_JUMP_TO)
         if has_patch and not is_qparent:
-            append(_('_rename'), self.rename_activated)
+            append(_('_Rename'), self.rename_activated, gtk.STOCK_EDIT)
         if has_applied and not is_qparent:
-            append(_('_finish applied'), self.finish_activated)
+            append(_('_Finish Applied'), self.finish_activated,
+                   gtk.STOCK_APPLY)
         if not is_applied and not is_qparent:
-            append(_('_delete'), self.delete_activated)
-            append(_('delete --keep'), self.delete_keep_activated)
+            append(_('_Delete'), self.delete_activated, gtk.STOCK_DELETE)
             if has_applied and not is_qparent:
-                append(_('f_old'), self.fold_activated)
-            if self.hasqup and not is_next:
-                append(_('make it _next'), self.mknext_activated)
+                append(_('F_old'), self.fold_activated, gtk.STOCK_DIRECTORY)
+            if self.get_num_unapplied() > 1:
+                sub = self.create_reorder_menu()
+                m.append_submenu(_('Reorder'), sub, gtk.STOCK_INDEX)
 
+        menu = m.build()
+        if len(menu.get_children()) > 0:
+            menu.show_all()
+            menu.popup(None, None, None, 0, 0)
+
+    def show_patches_cmenu(self):
+        sel = self.list.get_selection()
+        patches = []
+        for path in sel.get_selected_rows()[1]:
+            row = self.model[path]
+            if row[MQ_INDEX] in (INDEX_SEPARATOR, INDEX_QPARENT):
+                continue
+            patches.append(row[MQ_NAME])
+
+        incl_applied, incl_unapplied = False, False
+        for patch in patches:
+            if self.is_applied(patch):
+                incl_applied = True
+            else:
+                incl_unapplied = True
+
+        m = gtklib.MenuBuilder()
+        if incl_unapplied:
+            m.append(_('_Delete'), lambda *a: self.qdelete(patches),
+                     gtk.STOCK_DELETE)
+        if self.has_applied() and incl_unapplied:
+            m.append(_('F_old'), lambda *a: self.qfold(patches),
+                     gtk.STOCK_DIRECTORY)
+        if not incl_applied and incl_unapplied:
+            sub = self.create_reorder_menu()
+            m.append_submenu(_('Reorder'), sub, gtk.STOCK_INDEX)
+
+        menu = m.build()
         if len(menu.get_children()) > 0:
             menu.show_all()
             menu.popup(None, None, None, 0, 0)
 
     def create_view_menu(self):
-        menu = gtk.Menu()
-        def append(item=None, handler=None, check=False,
-                   active=False, sep=False):
-            if sep:
-                item = gtk.SeparatorMenuItem()
-            else:
-                if isinstance(item, str):
-                    if check:
-                        item = gtk.CheckMenuItem(item)
-                        item.set_active(active)
-                    else:
-                        item = gtk.MenuItem(item)
-                item.set_border_width(1)
-            if handler:
-                item.connect('activate', handler)
-            menu.append(item)
-            return item
+        self.vmenu = {}
+        m = gtklib.MenuBuilder()
+
         def colappend(label, col_idx, active=True):
             def handler(menuitem):
                 col = self.cols[col_idx]
                 col.set_visible(menuitem.get_active())
             propname = self.col_to_prop(col_idx)
-            item = append(label, handler, check=True, active=active)
+            item = m.append(label, handler, ascheck=True, check=active)
             self.vmenu[propname] = item
 
-        self.vmenu = {}
+        colappend(_('Index'), MQ_INDEX)
+        colappend(_('Status'), MQ_STATUS, active=False)
+        colappend(_('Summary'), MQ_SUMMARY, active=False)
 
-        colappend(_('Show index'), MQ_INDEX)
-        colappend(_('Show status'), MQ_STATUS, active=False)
-        colappend(_('Show name'), MQ_NAME)
-        colappend(_('Show summary'), MQ_SUMMARY, active=False)
-
-        append(sep=True)
+        m.append_sep()
 
         def enable_editable(item):
             self.cells[MQ_NAME].set_property('editable', item.get_active())
-        item = append(_('Enable editable cells'), enable_editable,
-                check=True, active=False)
+        item = m.append(_('Enable editable cells'), enable_editable,
+                        ascheck=True, check=False)
         self.vmenu['editable-cell'] = item
-        item = append(_("Show 'qparent'"), lambda item: self.refresh(),
-                check=True, active=True)
+        item = m.append(_("Show 'qparent'"), lambda item: self.refresh(),
+                        ascheck=True, check=True)
         self.vmenu['show-qparent'] = item
 
+        menu = m.build()
         menu.show_all()
         return menu
 
@@ -644,12 +863,14 @@ class MQWidget(gtk.VBox):
             self.cmd.set_result(_('Canceled'), style='error')
         else:
             self.cmd.set_result(_('Failed'), style='error')
-        self.repo.mq.invalidate()
+        hglib.invalidaterepo(self.repo)
         self.refresh()
         if not noemit:
             self.emit('repo-invalidated')
 
     def do_get_property(self, property):
+        if property.name == 'name-column-visible':
+            return True
         try:
             return self.vmenu[property.name].get_active()
         except:
@@ -677,6 +898,7 @@ class MQWidget(gtk.VBox):
     def list_pressed(self, list, event):
         x, y = int(event.x), int(event.y)
         pathinfo = list.get_path_at_pos(x, y)
+
         if event.button == 1:
             if not pathinfo:
                 # HACK: clear selection after this function calling,
@@ -685,9 +907,30 @@ class MQWidget(gtk.VBox):
                     selection = list.get_selection()
                     selection.unselect_all()
                 gtklib.idle_add_single_call(unselect)
-        elif event.button == 3:
-            if pathinfo:
-                self.show_patch_cmenu(self.list, pathinfo[0])
+
+        elif event.button == 3 and pathinfo:
+            sel = list.get_selection()
+            sel_rows = sel.get_selected_rows()[1] # list of paths
+            if pathinfo[0] not in sel_rows:
+                sel.unselect_all()
+                sel.select_path(pathinfo[0])
+            return True
+
+    def list_released(self, list, event):
+        if event.button != 3:
+            return
+
+        sel = list.get_selection()
+        count = sel.count_selected_rows()
+        if count == 1:
+            x, y = int(event.x), int(event.y)
+            pathinfo = list.get_path_at_pos(x, y)
+            if not pathinfo:
+                return
+            self.show_patch_cmenu(pathinfo[0])
+
+        elif 1 < count:
+            self.show_patches_cmenu()
 
     def list_sel_changed(self, list):
         path, focus = list.get_cursor()
@@ -698,7 +941,7 @@ class MQWidget(gtk.VBox):
         try:
             ctx = self.repo[patchname]
             revid = ctx.rev()
-        except hglib.RepoError:
+        except (error.RepoError, error.RepoLookupError, error.LookupError):
             revid = -1
         self.emit('patch-selected', revid, patchname)
 
@@ -709,17 +952,11 @@ class MQWidget(gtk.VBox):
         if self.mqloaded and self.has_applied():
             self.scroll_to_current()
 
-    def popall_clicked(self, toolbutton):
-        self.qpop(all=True)
-
-    def pop_clicked(self, toolbutton):
-        self.qpop()
-
-    def push_clicked(self, toolbutton):
-        self.qpush()
-
-    def pushall_clicked(self, toolbutton):
-        self.qpush(all=True)
+    def dnd_received(self, widget, context, x, y, sel, target, *args):
+        if target == MQ_DND_URI_LIST:
+            paths = gtklib.normalize_dnd_paths(sel.data)
+            if paths:
+                self.emit('files-dropped', paths, sel.data)
 
     ### context menu signal handlers ###
 
@@ -729,9 +966,6 @@ class MQWidget(gtk.VBox):
     def delete_activated(self, menuitem, row):
         self.qdelete(row[MQ_NAME])
 
-    def delete_keep_activated(self, menuitem, row):
-        self.qdelete(row[MQ_NAME], keep=True)
-    
     def rename_activated(self, menuitem, row):
         self.qrename_ui(row[MQ_NAME])
 
@@ -740,6 +974,3 @@ class MQWidget(gtk.VBox):
 
     def fold_activated(self, menuitem, row):
         self.qfold(row[MQ_NAME])
-
-    def mknext_activated(self, menuitem, row):
-        self.mknext(row[MQ_NAME])
