@@ -9,6 +9,7 @@
 import os
 import sys
 import time
+import threading
 from mercurial.i18n import _
 from mercurial import hg
 
@@ -18,14 +19,16 @@ def get_system_times():
         t = (t[0], t[1], t[2], t[3], time.clock())
     return t
 
-def version():
-    try:
-        import __version__
-        return __version__.version
-    except ImportError:
-        return _('unknown')
-
 if os.name == 'nt':
+    def browse_url(url):
+        def start_browser():
+            try:
+                import win32api
+                win32api.ShellExecute(0, 'open', url, None, None, 0)
+            except (ImportError, Exception):
+                pass
+        threading.Thread(target=start_browser).start()
+
     def shell_notify(paths):
         try:
             from win32com.shell import shell, shellcon
@@ -143,3 +146,22 @@ else:
 
     def update_thgstatus(*args, **kws):
         pass
+
+    def browse_url(url):
+        def start_browser():
+            if sys.platform == 'darwin':
+                # use Mac OS X internet config module (removed in Python 3.0)
+                import ic
+                ic.launchurl(url)
+            else:
+                try:
+                    import gconf
+                    client = gconf.client_get_default()
+                    browser = client.get_string(
+                            '/desktop/gnome/url-handlers/http/command') + '&'
+                    os.system(browser % url)
+                except ImportError:
+                    # If gconf is not found, fall back to old standard
+                    os.system('firefox ' + url)
+        threading.Thread(target=start_browser).start()
+
